@@ -1,9 +1,12 @@
 from ngutils.functions import *
 
-__version__ = '0.9.11'
+__version__ = '0.9.14'
 
 __all__ = [
     'accel_steps',
+    'flatten',
+    'hash_hd',
+    'read_files_contents',
     'read_urls_contents',
     'reduce_content',
     'tune_steps',
